/*
 * Copyright (C) 2022 European Spallation Source ERIC.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package org.openepics.names.rest.beans.element;

import java.io.Serializable;
import java.util.Date;
import java.util.UUID;

import org.openepics.names.rest.beans.Status;

import io.swagger.v3.oas.annotations.media.Schema;

/**
 * Bean (data transfer object) for communication and (json, xml) serialization.
 * Intended for operations that read names.
 *
 * @author Lars Johansson
 */
@Schema(description = "A collection of attributes that is used to encapsulate a name. "
        + "It used to show information for a name.")
public class NameElement extends NameElementCommand implements Serializable {

    /**
     *
     */
    private static final long serialVersionUID = 2716415801117244814L;

    @Schema(description = "Mnemonic path for for the system structure.")
    private String systemstructure;
    @Schema(description = "Mnemonic path for for the device structure.")
    private String devicestructure;
    @Schema(description = "Name (verbose) of the name entry.")
    private String name;
    @Schema(description = "Status of the name entry.")
    private Status status;
    @Schema(description = "If the name entry is latest (with status APPROVED) in its line of entries.")
    private Boolean latest;
    @Schema(description = "If the name entry is deleted.")
    private Boolean deleted;
    @Schema(description = "Date and time when the name entry was created.")
    private Date when;
    @Schema(description = "Name (user) of who created the name entry.")
    private String who;

    /**
     * Public constructor.
     */
    public NameElement() {
        super();
    }

    /**
     * Public constructor.
     *
     * @param uuid uuid
     * @param parentsystemstructure parent system structure uuid
     * @param parentdevicestructure parent device structure uuid
     * @param index index
     * @param description description
     * @param comment comment
     */
    public NameElement(
            UUID uuid, UUID parentsystemstructure, UUID parentdevicestructure, String index,
            String description, String comment) {
        super(uuid, parentsystemstructure, parentdevicestructure, index, description, comment);
    }

    /**
     * Public constructor.
     *
     * @param uuid uuid
     * @param parentsystemstructure parent system structure uuid
     * @param parentdevicestructure parent device structure uuid
     * @param index index
     * @param description description
     * @param comment comment
     * @param systemstructure system structure mnemonic path
     * @param devicestructure device structure mnemonic path
     * @param name name
     * @param status status
     * @param latest latest
     * @param deleted deleted
     * @param when when
     * @param who who
     */
    public NameElement(
            UUID uuid, UUID parentsystemstructure, UUID parentdevicestructure, String index,
            String description, String comment,
            String systemstructure, String devicestructure, String name,
            Status status, Boolean latest, Boolean deleted,
            Date when, String who) {
        super(uuid, parentsystemstructure, parentdevicestructure, index, description, comment);
        this.systemstructure = systemstructure;
        this.devicestructure = devicestructure;
        this.name = name;
        this.status = status;
        this.latest = latest;
        this.deleted = deleted;
        this.when = when;
        this.who = who;
    }

    public String getSystemstructure() {
        return systemstructure;
    }
    public void setSystemstructure(String systemstructure) {
        this.systemstructure = systemstructure;
    }
    public String getDevicestructure() {
        return devicestructure;
    }
    public void setDevicestructure(String devicestructure) {
        this.devicestructure = devicestructure;
    }
    public String getName() {
        return name;
    }
    public void setName(String name) {
        this.name = name;
    }
    public Status getStatus() {
        return status;
    }
    public void setStatus(Status status) {
        this.status = status;
    }
    public Boolean isLatest() {
        return latest;
    }
    public void setLatest(Boolean latest) {
        this.latest = latest;
    }
    public Boolean isDeleted() {
        return deleted;
    }
    public void setDeleted(Boolean deleted) {
        this.deleted = deleted;
    }
    public Date getWhen() {
        return when;
    }
    public void setWhen(Date when) {
        this.when = when;
    }
    public String getWho() {
        return who;
    }
    public void setWho(String who) {
        this.who = who;
    }

    @Override
    public boolean equals(Object obj) {
        if (this == obj)
            return true;
        if (obj == null)
            return false;
        if (getClass() != obj.getClass())
            return false;

        return equals ((NameElement) obj);
    }

    public boolean equals(NameElement other) {
        if (other == null)
            return false;

        if (!super.equals(other)) {
            return false;
        }

        if (getSystemstructure() == null) {
            if (other.getSystemstructure() != null)
                return false;
        } else if (!getSystemstructure().equals(other.getSystemstructure()))
            return false;
        if (getDevicestructure() == null) {
            if (other.getDevicestructure() != null)
                return false;
        } else if (!getDevicestructure().equals(other.getDevicestructure()))
            return false;
        if (getName() == null) {
            if (other.getName() != null)
                return false;
        } else if (!getName().equals(other.getName()))
            return false;
        if (getStatus() == null) {
            if (other.getStatus() != null)
                return false;
        } else if (!getStatus().equals(other.getStatus()))
            return false;
        if (isLatest() == null) {
            if (other.isLatest() != null)
                return false;
        } else if (!isLatest().equals(other.isLatest()))
            return false;
        if (isDeleted() == null) {
            if (other.isDeleted() != null)
                return false;
        } else if (!isDeleted().equals(other.isDeleted()))
            return false;
        if (getWhen() == null) {
            if (other.getWhen() != null)
                return false;
        } else if (!getWhen().equals(other.getWhen()))
            return false;
        if (getWho() == null) {
            if (other.getWho() != null)
                return false;
        } else if (!getWho().equals(other.getWho()))
            return false;

        return true;
    }

    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        sb.append("{");
        sb.append("\"uuid\": "                    + getUuid());
        sb.append(", \"parentsystemstructure\": " + getParentsystemstructure());
        sb.append(", \"parentdevicestructure\": " + getParentdevicestructure());
        sb.append(", \"systemstructure\": "       + getSystemstructure());
        sb.append(", \"devicestructure\": "       + getDevicestructure());
        sb.append(", \"index\": "                 + getIndex());
        sb.append(", \"name\": "                  + getName());
        sb.append(", \"description\": "           + getDescription());
        sb.append(", \"status\": "                + getStatus());
        sb.append(", \"latest\": "                + isLatest());
        sb.append(", \"deleted\": "               + isDeleted());
        sb.append(", \"when\": "                  + getWhen());
        sb.append(", \"who\": "                   + getWho());
        sb.append(", \"comment\": "               + getComment());
        sb.append("}");
        return sb.toString();
    }

    public String toStringSimple() {
        StringBuilder sb = new StringBuilder();
        sb.append("{");
        sb.append("\"uuid\": "                    + getUuid());
        sb.append(", \"name\": "                  + getName());
        sb.append(", \"status\": "                + getStatus());
        sb.append(", \"latest\": "                + isLatest());
        sb.append(", \"deleted\": "               + isDeleted());
        sb.append("}");
        return sb.toString();
    }

}
