/*
 * Copyright (C) 2022 European Spallation Source ERIC.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package org.openepics.names.rest.controller;

import org.apache.commons.lang3.StringUtils;
import org.openepics.names.rest.beans.response.Response;
import org.openepics.names.service.exception.DataConflictException;
import org.openepics.names.service.exception.DataDeletedException;
import org.openepics.names.service.exception.DataExistException;
import org.openepics.names.service.exception.DataNotAvailableException;
import org.openepics.names.service.exception.DataNotCorrectException;
import org.openepics.names.service.exception.DataNotFoundException;
import org.openepics.names.service.exception.DataNotValidException;
import org.openepics.names.service.exception.InputNotAvailableException;
import org.openepics.names.service.exception.InputNotCorrectException;
import org.openepics.names.service.exception.InputNotEmptyException;
import org.openepics.names.service.exception.InputNotValidException;
import org.openepics.names.service.exception.ServiceException;
import org.openepics.names.util.TextUtil;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.web.bind.annotation.ExceptionHandler;
import org.springframework.web.bind.annotation.RestControllerAdvice;
import org.springframework.web.context.request.WebRequest;
import org.springframework.web.servlet.mvc.method.annotation.ResponseEntityExceptionHandler;

/**
 * Global exception handler to ensure exceptions are communicated to request origin in a uniform way.
  * <br/><br/>
 * Note
 * <ul>
 * <li>400 {@link HttpStatus#BAD_REQUEST}</li>
 * <li>404 {@link HttpStatus#NOT_FOUND}</li>
 * <li>409 {@link HttpStatus#CONFLICT}</li>
 * <li>422 {@link HttpStatus#UNPROCESSABLE_ENTITY}</li>
 * <li>500 {@link HttpStatus#INTERNAL_SERVER_ERROR}</li>
 * </ul>
*
 * @author Lars Johansson
 */
@RestControllerAdvice
public class GlobalControllerExceptionHandler extends ResponseEntityExceptionHandler {

    // note
    //     no logging here

    @ExceptionHandler
    protected ResponseEntity<Response> handleConflict(RuntimeException ex, WebRequest request) {
        Response response = new Response("", "", "");
        response.setMessage(TextUtil.OPERATION_COULD_NOT_BE_PERFORMED);

        HttpStatus resultStatus = HttpStatus.INTERNAL_SERVER_ERROR;

        if (ex instanceof ServiceException se) {
            response.setMessage(StringUtils.trimToEmpty(ex.getMessage()));
            response.setDetails(StringUtils.trimToEmpty(se.getDetails()));
            response.setField(StringUtils.trimToEmpty(se.getField()));
            resultStatus = HttpStatus.BAD_REQUEST;

            if (ex instanceof InputNotAvailableException
                    || ex instanceof InputNotCorrectException
                    || ex instanceof InputNotEmptyException
                    || ex instanceof InputNotValidException) {
                resultStatus = HttpStatus.UNPROCESSABLE_ENTITY;
            }

            if (ex instanceof DataNotAvailableException || ex instanceof DataNotFoundException) {
                resultStatus = HttpStatus.NOT_FOUND;
            }

            if (ex instanceof DataNotCorrectException || ex instanceof DataNotValidException) {
                resultStatus = HttpStatus.UNPROCESSABLE_ENTITY;
            }

            if (ex instanceof DataDeletedException) {
                 resultStatus = HttpStatus.UNPROCESSABLE_ENTITY;
            }

            if (ex instanceof DataConflictException || ex instanceof DataExistException) {
                 resultStatus = HttpStatus.CONFLICT;
            }
        }

        return new ResponseEntity<>(response, Response.HEADER_JSON, resultStatus);
    }

}
