/*
 * Copyright (C) 2021 European Spallation Source ERIC.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package org.openepics.names.util;

import java.util.Date;
import java.util.UUID;

import org.openepics.names.repository.model.DeviceType;
import org.openepics.names.repository.model.Name;
import org.openepics.names.repository.model.Subsystem;
import org.openepics.names.repository.model.System;
import org.openepics.names.repository.model.SystemGroup;
import org.openepics.names.rest.beans.Status;
import org.openepics.names.rest.beans.element.NameElement;
import org.openepics.names.rest.beans.element.NameElementCommand;

/**
 * Utility class to assist in populating name elements based on repository content.
 * <br/><br/>
 * Different strategies for population of name elements are used in different methods.
 * The difference in strategies is for finding out system structure and device structure mnemonic paths.
 *
 * @author Lars Johansson
 */
public class NameElementUtil {

    // note
    //     handling of system structure, device structure
    //         parent system structure uuid (system group, system, subsystem) - ability to find structure
    //         parent device structure uuid (device type)

    /**
     * This class is not to be instantiated.
     */
    private NameElementUtil() {
        throw new IllegalStateException("Utility class");
    }

    /**
     * Populate and return name element for name.
     *
     * @param name name
     * @return name element
     */
    public static NameElement getNameElement(Name name) {
        if (name == null) {
            return null;
        }

        UUID parentsystemstructure = name.getSubsystemUuid() != null
                ? name.getSubsystemUuid()
                : name.getSystemUuid() != null
                        ? name.getSystemUuid()
                        : name.getSystemgroupUuid() != null
                                ? name.getSystemgroupUuid()
                                : null;

        return getNameElement(
                name.getUuid(),
                parentsystemstructure, name.getDevicetypeUuid(),
                NamingConventionUtil.extractMnemonicPathSystemStructure(name.getConventionName()),
                NamingConventionUtil.extractMnemonicPathDeviceStructure(name.getConventionName()),
                name.getInstanceIndex(), name.getConventionName(),
                name.getDescription(), Status.APPROVED, name.isLatest(), name.isDeleted(),
                name.getRequested(), name.getRequestedBy(), name.getRequestedComment());
    }

    /**
     * Populate and return name element.
     *
     * @param uuid uuid
     * @param parentsystemstructure parent system structure uuid (system group, system, subsystem)
     * @param parentdevicestructure parent device structure uuid (device type)
     * @param systemstructure system structure mnemonic path
     * @param devicestructure device structure mnemonic path
     * @param index instance index
     * @param name name
     * @param description description
     * @param status status
     * @param latest latest
     * @param deleted deleted
     * @param when when
     * @param who who
     * @param comment comment
     * @return name element
     */
    protected static NameElement getNameElement(
            UUID uuid,
            UUID parentsystemstructure, UUID parentdevicestructure,
            String systemstructure, String devicestructure,
            String index, String name,
            String description, Status status, Boolean latest, Boolean deleted,
            Date when, String who, String comment) {

        return new NameElement(
                uuid, parentsystemstructure, parentdevicestructure, index,
                description, comment,
                systemstructure, devicestructure, name,
                status, latest, deleted,
                when, who);
    }

    /**
     * Check if name element and name have same content.
     *
     * @param nameElement name element
     * @param name name
     * @param holderIRepositories holder repositories
     * @param holder holder repositories
     * @return true if name element and name have same content, false otherwise
     */
    public static boolean hasSameContent(NameElementCommand nameElement, Name name, HolderIRepositories holderIRepositories, HolderSystemDeviceStructure holder) {
        /*
           NameElement
               x parentsystemstructure (system group, system, subsystem)
               x parentdevicestructure (device type)
               x index
               x name
           NameElementCommand
               uuid
               parentsystemstructure
               parentdevicestructure
               index
               description
               comment
           Name
               x systemgroup_uuid
               x system_uuid
               x subsystem_uuid
               x devicetype_uuid
               x instance_index
               x convention_name
         */

        // find out system group, system, subsystem
        //     one of the three expected to be non-null, other two expected to be null
        SystemGroup systemGroup = holderIRepositories.getSystemGroupRepository().findLatestByUuid(nameElement.getParentsystemstructure().toString());
        System      system      = holderIRepositories.getSystemRepository().findLatestByUuid(nameElement.getParentsystemstructure().toString());
        Subsystem   subsystem   = holderIRepositories.getSubsystemRepository().findLatestByUuid(nameElement.getParentsystemstructure().toString());
        DeviceType  deviceType  = holderIRepositories.getDeviceTypeRepository().findLatestByUuid(nameElement.getParentdevicestructure().toString());

        String derivedName = null;
        if (systemGroup != null) {
            derivedName = NameUtil.getName(systemGroup, deviceType, nameElement.getIndex(), holder);
        } else if (system != null) {
            derivedName = NameUtil.getName(system, deviceType, nameElement.getIndex(), holder);
        } else if (subsystem != null) {
            derivedName = NameUtil.getName(subsystem, deviceType, nameElement.getIndex(), holder);
        } else {
            return false;
        }

        if (nameElement.getUuid() == null) {
            if (name.getUuid() != null)
                return false;
        } else if (!nameElement.getUuid().equals(name.getUuid()))
            return false;
        if (derivedName == null) {
            if (name.getConventionName() != null)
                return false;
        } else if (!derivedName.equals(name.getConventionName()))
            return false;
        if (nameElement.getDescription() == null) {
            if (name.getDescription() != null)
                return false;
        } else if (!nameElement.getDescription().equals(name.getDescription()))
            return false;
        // no need to compare comment, it is comment to action that is taken, not considered state

        return true;
    }

}
